/*    
(C) 2007,2008 Michael Chaten
This file is part of GoodGame.
GoodGame is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 3 of the License
GoodGame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.goodgame.server.user;

import net.sf.goodgame.shared.UserInfoElement;
import java.nio.channels.SocketChannel;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.Semaphore;
import net.sf.goodgame.server.lobby.Lobby;
import net.sf.goodgame.server.game.Game;
import net.sf.goodgame.server.game.GameType;
import net.sf.goodgame.server.networking.IOManager;
import net.sf.goodgame.server.networking.ServerToClientLogic;
import net.sf.goodgame.shared.UserType;
import net.sf.goodgame.shared.network.NetworkMessage;
import net.sf.goodgame.server.persistence.Database;
import net.sf.goodgame.server.persistence.DatabaseManager;

/**
 *
 * @author Mike
 */
public class User implements Comparable {

    @Override
    public int compareTo(Object o) {
        return o.toString().compareTo(currentUsername);
    }
    private final static Map<String, User> users = new HashMap<String, User>();
    public final SocketChannel socket;
    private String currentUsername;
    private boolean initialized;
    private UserType type;
    private Collection<UserInfoElement> myInfo;
    private Lobby currentLobby;
    private GameType currentGameType;
    private Game currentGame;
    private FriendsList friendsList;
    private Collection<String> friendsOf;
    private final Semaphore instructionSemaphore;

    public User(SocketChannel socket) {
        assert DatabaseManager.getDatabase() != null;
        assert IOManager.getIO() != null;
        this.socket = socket;
        initialized = false;
        type = null;
        myInfo = new Vector<UserInfoElement>();
        friendsList = null;
        friendsOf = null;
        instructionSemaphore = new Semaphore(1, true);
    }

    public static User getUser(String username) {
        synchronized (users) {
            return users.get(username);
        }
    }

    public Collection<UserInfoElement> getUserInfo() {
        return myInfo;
    }

    public void register(String username, String password, String email) {
        final byte MAX_NAME_LEN = 26;
        final byte MAX_PASS_LEN = 26;
        //Todo: Check Email for validity.
        if (username.length() > MAX_NAME_LEN ||
                password.length() > MAX_PASS_LEN) {
            this.send(ServerToClientLogic.loginFailed());
        } else if (DatabaseManager.getDatabase().registerUser(username, password, email, UserType.regular)) {
            login(username, password);
        } else {
            this.send(ServerToClientLogic.loginFailed());
        }
    }

    public boolean isInitialized() {
        return initialized;
    }

    public String getUsername() {
        return currentUsername;
    }

    public UserType getType() {
        return type;
    }

    public boolean isLoggedIn() {
        return type != null;
    }

    public void disconnect() {
        DatabaseManager.getDatabase().logout(currentUsername);

        if (currentLobby != null) {
            currentLobby.removeUser(this);
        }
        if (currentGame != null) {
            currentGame.removeUser(this);
        }
        Lobby.removeLobbyListener(this);
        GameType.gameTypeList.removeListener(this);
        IOManager.getIO().userDisconnected(this);
    }

    public void initialized() {
        //Todo: Change this to something more friendly. x
        users.put(currentUsername, this);
        initialized = true;
    }
    /*
     * @return True if the login was successful\
     * False otherwise
     */

    public void login(String name, String password) {
        Database db = DatabaseManager.getDatabase();
        if (db.login(name, password)) {
            synchronized (users) {
                type = db.getType(name);
                users.remove(currentUsername);
                currentUsername = name;
                myInfo = db.getUserInfo(currentUsername);
                users.put(name, this);
                friendsList = db.getFriendsList(this);
                friendsOf = db.getFriendsOfList(this.getUsername());
                for (String friendOf : friendsOf) {
                    User user = User.getUser(friendOf);
                    if (user == null) {
                        continue;
                    }
                    user.friendLoggedIn(this);
                }
                send(ServerToClientLogic.loginOk(this));
                friendsList.initialize();
                Lobby.addLobbyListener(this);
                GameType.gameTypeList.addListener(this);
                this.switchLobby(Lobby.getLobby((short) 0));
                
            }
        } else {
            send(ServerToClientLogic.loginFailed());
        }
    }

    public void logout() {
        synchronized (users) {
            for (String friendOf : friendsOf) {
                User user = User.getUser(friendOf);
                if (user == null) {
                    continue;
                }
                user.friendLoggedOut(this);
            }
            DatabaseManager.getDatabase().logout(currentUsername);

            users.remove(currentUsername);
            Lobby.removeLobbyListener(this);
            GameType.gameTypeList.removeListener(this);
            
            leaveLobby();
            leaveGame();
            send(ServerToClientLogic.logout(this));
            currentUsername = null;
            type = null;
            friendsList = null;
            
        }
    }

    public void switchLobby(Lobby lobby) {
        if (currentLobby != null) {
            currentLobby.removeUser(this);
        }
        currentLobby = lobby;
        send(ServerToClientLogic.switchLobby(lobby));
        currentLobby.addUser(this);
    }

    public void switchGameType(GameType gt) {
        if (gt == currentGameType) {
            return;
        }
        this.send(ServerToClientLogic.switchGameType(gt));
        currentLobby.changeUserGameType(this, gt);
        currentGameType = gt;
    }

    public GameType getCurrentGameType() {
        return currentGameType;
    }

    public void joinGame(short gameID) {
        if (currentGame != null) {
            return;
        } else {
            currentGame = currentLobby.addUserToGame(this, gameID);
        }
    }

    public void makeGame() {
        if (currentGame != null) {
            leaveGame();
        }
        currentGame = currentLobby.addUserToGame(this, currentLobby.makeNewGame(this));
    }

    public void leaveGame() {
        if (currentGame != null) {
            send(ServerToClientLogic.leaveGame());
            currentLobby.removeUserFromGame(this, currentGame);
            currentGame = null;
        }
    }
    public void leaveLobby() {
        if(currentLobby != null) {
            currentLobby.removeUser(this);
            currentLobby = null;
        }
    }

    public void send(NetworkMessage message) {
        if (socket.isConnected()) {
            IOManager.getIO().send(message, this);
        }
    }

    public Lobby getCurrentLobby() {
        return currentLobby;
    }

    public Game getCurrentGame() {
        return currentGame;
    }

    public void instructionExecuting() {
        instructionSemaphore.acquireUninterruptibly();
    }

    public void instructionFinished() {
        instructionSemaphore.release();
    }

    @Override
    public String toString() {
        return currentUsername;
    }

    public void addFriend(String username) {
        if (username.equalsIgnoreCase(getUsername())) {
            friendsList.addFriend(username);
            DatabaseManager.getDatabase().addFriend(this, username);
        }
    }

    public void removeFriend(String username) {
        friendsList.removeFriend(username);
        DatabaseManager.getDatabase().removeFriend(this, username);
    }

    public void friendLoggedIn(User friend) {
        friendsList.friendLoggedIn(friend);
    }

    public void friendLoggedOut(User friend) {
        friendsList.friendLoggedOut(friend);
    }

    public void addFriendOf(String user) {
        friendsOf.add(user);
    }

    public void removeFriendOf(String user) {
        friendsOf.remove(user);
    }
}
